import os
import jinja2
import sendgrid
from sendgrid.helpers.mail import Email, Content, Mail

TEMPLATE_DIR = os.path.join(os.path.dirname(os.path.realpath(__file__)), "templates")

class AviSendGridClient(sendgrid.SendGridAPIClient):

    def __init__(self, apikey):
        super(AviSendGridClient, self).__init__(apikey=apikey)
        self.jinja2_env = jinja2.Environment(
            loader=jinja2.FileSystemLoader(TEMPLATE_DIR),
            trim_blocks=True,
        )

    def send_activation_email(self, from_email, to_email, first_name, activation_link, images_uri):
        variables = {
            'first_name': first_name,
            'username': to_email,
            'activation_link': activation_link,
            'images_uri': images_uri,
        }
        template = self.jinja2_env.get_template("activation_email.template")
        from_email = Email(from_email)
        to_email = Email(to_email)
        subject = "Avi Networks SaaS account activation"
        content = Content("text/html", template.render(variables))
        mail = Mail(from_email, subject, to_email, content)
        response = self.client.mail.send.post(request_body=mail.get())
        return response.status_code



API_VALIDATION_FAILED = 'E0000001'
AUTHENTICATION_FAILED = 'E0000004'
INVALID_SESSION = 'E0000005'
PERMISSION_DENIED = 'E0000006'
NOT_FOUND = 'E0000007'
INVALID_TOKEN = 'E0000011'
USER_ALREADY_ACTIVATED = 'E0000016'
ROLE_ALREADY_ASSIGNED = 'E0000090'

class Error(Exception):
    def __init__(self, error):
        if error is None:
            error = {}

        Exception.__init__(self, error.get('errorSummary'))

        self.error = error
        self.causes = error.get('errorCauses')
        self.code = error.get('errorCode')
        self.id = error.get('errorId')
        self.link = error.get('errorLink')
        self.summary = error.get('errorSummary')

    def __repr__(self):
        return str(self.error)


class NotFound(Error):
    def __init__(self, error):
        super(NotFound, self).__init__(error)


class RoleAlreadyAssigned(Error):
    def __init__(self, error):
        super(RoleAlreadyAssigned, self).__init__(error)


class UserAlreadyActivated(Error):
    def __init__(self, error):
        super(UserAlreadyActivated, self).__init__(error)


def factory(error):
    if error is None or not error.get('errorCode'):
        return Error(error)
    ec = error['errorCode']
    if ec == NOT_FOUND:
        return NotFound(error)
    if ec == ROLE_ALREADY_ASSIGNED:
        return RoleAlreadyAssigned(error)
    if ec == USER_ALREADY_ACTIVATED:
        return UserAlreadyActivated(error)
    return Error(error)



import re
import time
import json
import backoff
import requests
from requests.exceptions import RequestException
from datetime import datetime
import avi_okta.OktaError as OktaError

API_PATH = '/api/v1/'
INTERNAL_API_PATH = '/api/internal/'
HOME_PATH = '/app/UserHome'
SESSION_COOKIE_PATH = '/login/sessionCookieRedirect'


def _failure_code(rsp):
    return rsp.status_code < 200 or rsp.status_code >= 300


def _fatal_code(e):
    if not e.response:
        return False
    return 400 <= e.response.status_code < 500


def _giveup_handler(details):
    rsp = details['value']
    raise OktaError.factory(json.loads(rsp.text))


class ApiClient(object):

    def __init__(self, base_url, token=None, username=None, password=None):
        if not token and not (username and password):
            raise ValueError('An API token or username and password are required')

        self.base_url = base_url
        self.api_url = base_url + API_PATH

        self.sessions = {}
        self.session = requests.Session()
        self.session.headers.update({
            'Accept': 'application/json, application/xml',
            'Accept-Encoding': 'gzip, deflate, br',
            'Content-Type': 'application/json',
        })

        if token:
            self.session.headers.update({'Authorization': 'SSWS ' + token})
            self.sessions['token'] = self.session
            return

        # Authenticate to get a session token
        data = {
            'username':username,
            'password':password
        }
        rsp = self.post('authn', data=data)
        session_token = rsp.json()['sessionToken']

        # Trade session token for a session cookie and CSRF token
        params = {
            'token': session_token,
            'redirectUrl': base_url + HOME_PATH,
        }
        rsp = self._get(base_url + SESSION_COOKIE_PATH, params=params)
        match = re.search(r'xsrfToken">(.+?)<', rsp.content)
        csrf_token = match.group(1)
        self.session.headers.update({"X-Okta-XsrfToken": csrf_token})
        self.sessions['cookie'] = self.session

    def _init_internal_api_session(self):
        session = requests.Session()
        session.headers.update(self.session.headers)
        session.cookies.update(self.session.cookies)

        # Get a session token
        session.headers.update({"Accept": "*/*"})
        rsp = session.get(self.base_url + "/app/admin/sso/saml")
        match = re.search(r'token":\["(.+?)"', rsp.content)
        session_token = match.group(1)

        # Trade session token for a session cookie and CSRF token
        parts = self.base_url.split('.', 1)
        self.base_admin_url = "%s-admin.%s" % (parts[0], parts[1])
        session.cookies.clear()
        session.headers.update({
            'Content-Type': 'application/x-www-form-urlencoded',
        })
        data = {"token": session_token}
        rsp = session.post(self.base_admin_url + "/admin/sso/request", data=data)
        match = re.search(r'xsrfToken">(.+?)<', rsp.content)
        csrf_token = match.group(1)

        # Reset headers to standard values
        session.headers.update({
            'Accept': 'application/json, application/xml',
            'Content-Type': 'application/json',
            'X-Okta-XsrfToken': csrf_token,
        })
        self.sessions['cookie-internal'] = session

    def target_standard_api(self):
        self.session = self.sessions['cookie']
        self.api_url = self.base_url + API_PATH

    def target_internal_api(self):
        if not self.sessions.get('cookie-internal'):
            self._init_internal_api_session()
        self.session = self.sessions['cookie-internal']
        self.api_url = self.base_admin_url + INTERNAL_API_PATH

    def close_sessions(self):
        if not self.sessions.get('cookie'):
            return
        self.target_standard_api()
        for s in self.sessions.values():
            session_id = s.cookies.get('sid')
            try:
                self.delete("sessions/" + session_id)
            except Exception:
                pass

    @backoff.on_predicate(backoff.expo, _failure_code,
        max_tries=3,
        on_giveup=_giveup_handler
    )
    @backoff.on_exception(backoff.expo, RequestException,
        max_tries=3,
        giveup=_fatal_code
    )
    def get(self, path, params=None):
        url = self.api_url + path
        params = self._format_params(params)
        return self.session.get(url, params=params)


    @backoff.on_predicate(backoff.expo, _failure_code,
        max_tries=3,
        on_giveup=_giveup_handler
    )
    @backoff.on_exception(backoff.expo, RequestException,
        max_tries=3,
        giveup=_fatal_code
    )
    def post(self, path, data=None, params=None):
        url = self.api_url + path
        params = self._format_params(params)
        if data:
            data = json.dumps(data, cls=Serializer, separators=(',', ':'))
        return self.session.post(url, params=params, data=data)


    @backoff.on_predicate(backoff.expo, _failure_code,
        max_tries=3,
        on_giveup=_giveup_handler
    )
    @backoff.on_exception(backoff.expo, RequestException,
        max_tries=3,
        giveup=_fatal_code
    )
    def put(self, path, data=None, params=None):
        url = self.api_url + path
        params = self._format_params(params)
        if data:
            data = json.dumps(data, cls=Serializer)
        return self.session.put(url, params=params, data=data)


    @backoff.on_predicate(backoff.expo, _failure_code,
        max_tries=3,
        on_giveup=_giveup_handler
    )
    @backoff.on_exception(backoff.expo, RequestException,
        max_tries=3,
        giveup=_fatal_code
    )
    def delete(self, path, params=None):
        url = self.api_url + path
        params = self._format_params(params)
        return self.session.delete(url, params=params)

    @staticmethod
    def _format_params(params):
        if not params:
            return None
        formatted = {}
        for key, val in params.iteritems():
            if val is None:
                continue
            if type(val) == bool:
                val = str(val).lower()
            formatted[key] = val
        return formatted


class Serializer(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime):
            return obj.strftime('dt(%Y-%m-%dT%H:%M:%SZ)')
        elif isinstance(obj, object):
            no_nulls = self._remove_nulls(obj.__dict__)
            formatted = self._replace_alt_names(obj, no_nulls)
            return formatted
        else:
            return json.JSONEncoder.default(self, obj)

    def _remove_nulls(self, d):
        built = {}
        for k, v in d.iteritems():
            if v is None:
                continue

            if isinstance(v, dict):
                built[k] = self._remove_nulls(v)

            if isinstance(v, object) and hasattr(v, '__dict__'):
                built[k] = self._remove_nulls(v.__dict__)

            else:
                built[k] = v
        return built

    def _replace_alt_names(self, d):
        built = d.copy()
        if hasattr(obj, 'alt_names'):
            for key, value in obj.alt_names.iteritems():
                if value in built:
                    built[key] = built[value]
                    del built[value]
        return built


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="avi_okta",
    version="0.0.7",
    author="Neel Parikh",
    author_email="nparikh@avinetworks.com",
    description="Avi Networks Okta sdk",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/avinetworks/avi-internal",
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=[
        'requests',
        'backoff',
        'jinja2',
        'sendgrid',
    ],
    classifiers=[
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)




import os
import time
import json
import jinja2
import avi_okta.OktaError as OktaError
from avi_okta.ApiClient import ApiClient

TEMPLATE_DIR = os.path.join(os.path.dirname(os.path.realpath(__file__)), "templates")
BASE_URL = "https://avinetworks.oktapreview.com"


class AviOktaClient(ApiClient):
    def __init__(self, token=None, username=None, password=None):
        super(AviOktaClient, self).__init__(
            base_url=BASE_URL,
            token=token,
            username=username,
            password=password
        )
        self.jinja2_env = jinja2.Environment(
            loader=jinja2.FileSystemLoader(TEMPLATE_DIR),
            trim_blocks=True,
        )

    ### Applications ###
    def create_app(self, label, subdomain=None):
        try:
            app = self.get_app(label=label)
            app['_created'] = False
            return app
        except OktaError.NotFound:
            pass

        template = self.jinja2_env.get_template("app.template")
        variables = {
            "label": label,
            "subdomain": subdomain or label,
        }
        data = json.loads(template.render(variables))
        rsp = self.post('apps', data=data)
        app = rsp.json()
        app['_created'] = True

        template = self.jinja2_env.get_template("app_schema.template")
        data = json.loads(template.render())
        rsp = self.post('meta/schemas/apps/%s/default' % app['id'], data=data)
        return app

    def get_app(self, app_id=None, label=None):
        if app_id:
            rsp = self.get('apps/%s' % app_id)
            return rsp.json()
        apps = self.list_apps()
        for app in apps:
            if app['label'] == label:
                return app
        raise OktaError.factory({
            "errorCode": OktaError.NOT_FOUND,
            "errorSummary": "Resource not found: %s (AppInstance)" % label
        })

    def get_app_schema(self, app_id):
        rsp = self.get('meta/schemas/apps/%s/default' % app_id)
        return rsp.json()

    def get_app_metadata(self, app_id):
        rsp = self.get('apps/%s/sso/saml/metadata' % app_id)
        return rsp.content

    def list_app_users(self, app_id):
        rsp = self.get('apps/%s/users' % app_id)
        return rsp.json()

    def list_apps(self):
        rsp = self.get('apps')
        return rsp.json()

    def delete_app(self, app_id):
        self.deactivate_app(app_id)
        rsp = self.delete('apps/%s' % app_id)
        return rsp.status_code

    def deactivate_app(self, app_id):
        rsp = self.post('apps/%s/lifecycle/deactivate' % app_id)
        return rsp.status_code

    def assign_group_to_app(self, app_id, group_id):
        rsp = self.put('apps/%s/groups/%s' % (app_id, group_id))
        return rsp.json()

    def assign_user_to_app(self, app_id, user_id, attributes={}):
        user_login = self.get_user(user_id=user_id)['profile']['login']
        template = self.jinja2_env.get_template("app_user.template")
        variables = {
            "user_id": user_id,
            "username": user_login,
            "attributes": attributes,
        }
        data = json.loads(template.render(variables))
        rsp = self.post('apps/%s/users' % app_id, data=data)
        return rsp.json()

    def update_app_user_profile(self, app_id, user_id, attributes={}):
        template = self.jinja2_env.get_template("app_user.template")
        variables = {
            "attributes": attributes,
        }
        data = json.loads(template.render(variables))
        rsp = self.post('apps/%s/users/%s' % (app_id, user_id), data=data)
        return rsp.json()

    ### Users ###
    def create_user(self, first_name, last_name, email, organization=None,
            password=None, group_ids=[], activate=True):
        try:
            user = self.get_user(login=email)
            if group_ids:
                current_groups = self.get_user_groups(user['id'])
                current_group_ids = [g['id'] for g in current_groups]
                for group_id in group_ids:
                    if group_id not in current_group_ids:
                        self.assign_user_to_group(user['id'], group_id)
            user['_created'] = False
            return user
        except OktaError.NotFound:
            pass

        template = self.jinja2_env.get_template("user.template")
        variables = {
            'first_name': first_name,
            'last_name': last_name,
            'email': email,
            'organization': organization,
            'password': password,
            'group_ids': group_ids,
        }
        data = json.loads(template.render(variables))
        rsp = self.post('users', data=data, params={'activate': activate})
        user = rsp.json()
        user['_created'] = True
        return user

    def get_user(self, user_id=None, login=None):
        identifier = user_id or login
        rsp = self.get('users/%s' % identifier)
        return rsp.json()

    def get_user_groups(self, user_id):
        rsp = self.get('users/%s/groups' % user_id)
        return rsp.json()

    def list_users(self):
        rsp = self.get('users')
        return rsp.json()

    def delete_user(self, user_id):
        self.deactivate_user(user_id=user_id)
        rsp = self.delete('users/%s' % user_id)
        return rsp.status_code

    def activate_user(self, user_id, send_email=False):
        params = {'sendEmail': send_email}
        try:
            rsp = self.post('users/%s/lifecycle/activate' % user_id, params=params)
            return rsp.json()
        except OktaError.UserAlreadyActivated:
            return None

    def deactivate_user(self, user_id):
        rsp = self.post('users/%s/lifecycle/deactivate' % user_id)
        return rsp.status_code

    def expire_user_password(self, user_id):
        rsp = self.post('users/%s/lifecycle/expire_password' % user_id)
        return rsp.status_code

    def assign_group_admin_role_to_user(self, user_id, group_id):
        try:
            data = {'type': 'USER_ADMIN'}
            rsp = self.post('users/%s/roles' % user_id, data=data)
            role = rsp.json()
        except OktaError.RoleAlreadyAssigned:
            rsp = self.get('users/%s/roles' % user_id)
            roles = rsp.json()
            for r in roles:
                if r['type'] == 'USER_ADMIN':
                    role = r
        rsp = self.put('users/%s/roles/%s/targets/groups/%s' %
                (user_id, role['id'], group_id))
        return rsp.status_code

    ### Groups ###
    def create_group(self, name):
        try:
            group = self.get_group(name=name)
            group['_created'] = False
            return group
        except OktaError.NotFound:
            pass

        template = self.jinja2_env.get_template("group.template")
        variables = {
            'name': name,
        }
        data = json.loads(template.render(variables))
        rsp = self.post('groups', data=data)
        group = rsp.json()
        group['_created'] = True
        return group

    def get_group(self, group_id=None, name=None):
        if group_id:
            rsp = self.get('groups/%s' % group_id)
            return rsp.json()
        rsp = self.get('groups', params={'q': name})
        groups = rsp.json()
        if groups:
            return groups[0]
        raise OktaError.factory({
            "errorCode": OktaError.NOT_FOUND,
            "errorSummary": "Resource not found: %s (UserGroup)" % name
        })

    def get_group_users(self, group_id):
        rsp = self.get("groups/%s/users" % group_id)
        return rsp.json()

    def list_groups(self):
        rsp = self.get('groups')
        return rsp.json()

    def delete_group(self, group_id):
        rsp = self.delete('groups/%s' % group_id)
        return rsp.status_code

    def assign_user_to_group(self, user_id, group_id):
        rsp = self.put('groups/%s/users/%s' % (group_id, user_id))
        return rsp.status_code

    ### Trusted Origins ###
    def create_trusted_origin(self, name, origin, cors=True, redirect=True):
        try:
            trusted_origin = self.get_trusted_origin(name=name)
            trusted_origin['_created'] = False
            return trusted_origin
        except OktaError.NotFound:
            pass

        scopes = []
        if cors:
            scopes.append('CORS')
        if redirect:
            scopes.append('REDIRECT')
        template = self.jinja2_env.get_template("trusted_origin.template")
        variables = {
            'name': name,
            'origin': origin,
            'scopes': scopes,
        }
        data = json.loads(template.render(variables))
        rsp = self.post('trustedOrigins', data=data)
        trusted_origin = rsp.json()
        trusted_origin['_created'] = True
        return trusted_origin

    def get_trusted_origin(self, trusted_origin_id=None, name=None):
        if trusted_origin_id:
            rsp = self.get('trustedOrigins/%s' % trusted_origin_id)
            return rsp.json()
        trusted_origins = self.list_trusted_origins()
        for trusted_origin in trusted_origins:
            if trusted_origin['name'] == name:
                return trusted_origin
        raise OktaError.factory({
            "errorCode": OktaError.NOT_FOUND,
            "errorSummary": "Resource not found: %s (TrustedOrigin)" % name
        })

    def list_trusted_origins(self):
        rsp = self.get('trustedOrigins')
        return rsp.json()

    def delete_trusted_origin(self, trusted_origin_id):
        self.deactivate_trusted_origin(trusted_origin_id)
        rsp = self.delete('trustedOrigins/%s' % trusted_origin_id)
        return rsp.status_code

    def deactivate_trusted_origin(self, trusted_origin_id):
        rsp = self.post('trustedOrigins/%s/lifecycle/deactivate' % trusted_origin_id)
        return rsp.status_code

    ### API Tokens ###
    def create_token(self, name):
        try:
            token = self.get_token(name=name)
            token['_created'] = False
            return token
        except OktaError.NotFound:
            pass

        try:
            self.target_internal_api()
            rsp = self.post('tokens', data={'name': name})
            token = rsp.json()
            token['_created'] = True
            return token
        finally:
            self.target_standard_api()

    def get_token(self, token_id=None, name=None):
        try:
            self.target_internal_api()
            if token_id:
                rsp = self.get('tokens/%s' % token_id)
                return rsp.json()
            tokens = self.list_tokens()
            for token in tokens:
                if token['name'] == name:
                    return token
            raise OktaError.factory({
                "errorCode": OktaError.NOT_FOUND,
                "errorSummary": "Resource not found: %s (Token)" % name
            })
        finally:
            self.target_standard_api()

    def list_tokens(self):
        try:
            self.target_internal_api()
            rsp = self.get('tokens')
            return rsp.json()
        finally:
            self.target_standard_api()

    def delete_token(self, token_id):
        try:
            self.target_internal_api()
            rsp = self.post("tokens/%s/revoke" % token_id)
            return rsp.status_code
        finally:
            self.target_standard_api()
